/*********************************************************************

  RuleBook.java	(C) Kenneth J. Turner, Mark A. Jennings	25/11/01 V1.3

  This class defines an object that serves the RulesResolver object by
  providing a method to return a specific rule object. This class also
  constructs all the Rules on creation.

  KJT 14/08/01:

    o Minor tidy-up of code

    o Minor reformulation of rules

    o Minor correction of rule flags

    o Rule handling change to array of indexes

  KJT 25/11/01:

    o Updated for 2001-2004 rules

    o Rule 18.3 split

*********************************************************************/

import waba.ui.*;
import waba.fx.*;
import waba.sys.*;

public class RuleBook {
  public static final int NUM_RULES = 13;
  private static Rule[] theRules = new Rule[NUM_RULES];
  private int tempRules[] = new int[NUM_RULES];
  boolean setFlags[] = new boolean[Rule.NUM_FLAGS];
  private boolean tempFlags[] = new boolean[Rule.NUM_FLAGS];
  private String code, text;
  private int proximityRule = 5;

  // basic constructor

  public RuleBook () {
    clearFlags ();

    // on opposite tacks
    code ="10";
    text =
      "When boats are on opposite tacks,\n" +
      "a port-tack boat shall keep clear of\n" +
      "a starboard-tack boat.";
    setFlags[Rule.OPPOSITE_TACKS] = true;
    setFlags[Rule.STARBOARD_TACK] = true;
    theRules[1] = new Rule (code, text, setFlags);
    clearFlags ();

    // on the same tack, overlapped
    code = "11";
    text =
      "When boats are on the same tack\n" +
      "and overlapped, a windward boat\n" +
      "shall keep clear of a leeward boat.";
    setFlags[Rule.OPPOSITE_TACKS] = false;
    setFlags[Rule.STARBOARD_TACK] = false;
    setFlags[Rule.OVERLAPPED] = true;
    setFlags[Rule.CLEAR_AHEAD] = false;
    setFlags[Rule.TO_LEEWARD] = true;
    theRules[2] = new Rule (code, text, setFlags);
    clearFlags ();

    // on the same tack, not overlapped
    code = "12";
    text =
      "When boats are on the same tack\n" +
      "and not overlapped, a boat clear\n" +
      "astern shall keep clear of a boat\n" +
      "clear ahead.";
    setFlags[Rule.CLEAR_AHEAD] = true;
    theRules[3] = new Rule (code, text, setFlags);
    clearFlags ();

    // while tacking
    code = "13";
    text =
      "After a boat passes head to wind,\n" +
      "she shall keep clear of other boats\n" +
      "until she is on a close-hauled course.\n" +
      "During that time Rules 10, 11 and 12\n" +
      "do not apply.  If two boats are\n" +
      "subject to this rule at the same time\n" +
      "the one on the other's port side shall\n" +
      "keep clear.";
    setFlags[Rule.CLEAR_AHEAD] = true;
    setFlags[Rule.TACKING_TO_PORT] = true;
    setFlags[Rule.TACKING_TO_STARBOARD] = true;
    setFlags[Rule.HEAD_TO_WIND] = true;
    theRules[4] = new Rule (code, text, setFlags);
    clearFlags ();

    // avoiding contact
    code = "14";
    text =
      "A boat shall avoid contact with\n" +
      "another boat if reasonably possible.\n" +
      "However, a right-of-way boat or\n" +
      "one entitled to room (a) need not\n" +
      "act to avoid contact until it is clear\n" +
      "that the other boat is not keeping\n" +
      "clear or giving room, and (b) shall\n" +
      "not be penalized under this rule\n" +
      "unless there is contact that causes\n" +
      "damage.";
    setFlags[Rule.PROXIMITY] = true;
    theRules[5] = new Rule (code, text, setFlags);
    clearFlags ();

    // same tack proper course (1)
    code = "17.1";
    text =
      "If a boat clear astern becomes\n" +
      "overlapped within two of her\n" +
      "hull lengths to leeward of a\n" +
      "boat on the same tack, she\n" +
      "shall not sail above her proper\n" +
      "course while they remain\n" +
      "overlapped within that distance,\n" +
      "unless in doing so she promptly\n" +
      "sails astern of the other boat.\n" +
      "This rule does not apply if the\n" +
      "overlap begins while the windward\n" +
      "boat is required by Rule 13 to\n" +
      "keep clear.";
    setFlags[Rule.ABOVE_PROPER_COURSE] = true;
    setFlags[Rule.INSIDE_TWO_LENGTHS] = true;
    setFlags[Rule.OVERLAPPED] = true;
    setFlags[Rule.TO_LEEWARD] = true;
    theRules[6] = new Rule (code, text, setFlags);
    clearFlags ();

    // same tack proper course (2)
    code = "17.2";
    text =
      "Except on a beat to windward, while\n" +
      "a boat is less than two of her hull\n" +
      "lengths from a leeward boat or a\n" +
      "boat clear astern steering a course\n" +
      "to leeward of her, she shall not sail\n" +
      "below her proper course unless she\n" +
      "gybes.";
    setFlags[Rule.BELOW_PROPER_COURSE] = true;
    setFlags[Rule.INSIDE_TWO_LENGTHS] = true;
    theRules[7] = new Rule (code, text, setFlags);
    clearFlags ();

    // passing marks and obstructions: giving room and keeping clear
    code = "18.2(a)";
    text =
      "When boats are overlapped the\n" +
      "outside boat shall give the inside\n" +
      "boat room to pass the mark or\n" +
      "obstruction, and if the inside boat\n" +
      "has right of way the outside boat\n" +
      "shall keep clear. Other parts of\n" +
      "Rule 18 contain exceptions to this\n" +
      "rule.";
    setFlags[Rule.MARK_RULES] = true;
    setFlags[Rule.INSIDE_TWO_LENGTHS] = true;
    setFlags[Rule.OVERLAPPED] = true;
    theRules[8] = new Rule (code, text, setFlags);
    clearFlags ();

    // passing marks and obstructions: overlapped at the zone
    code = "18.2(b)";
    text =
      "If boats were overlapped before\n" +
      "either of them reached the two-\n" +
      "length zone and the overlap is\n" +
      "broken after one of them has\n" +
      "reached it, the boat that was on\n" +
      "the outside shall continue to\n" +
      "give the other boat room. If the\n" +
      "outside boat becomes clear astern\n" +
      "or overlapped inside the other\n" +
      "boat, she is not entitled to room\n" +
      "and shall keep clear.";
    setFlags[Rule.MARK_RULES] = true;
    setFlags[Rule.INSIDE_TWO_LENGTHS] = true;
    setFlags[Rule.OVERLAPPED] = true;
    setFlags[Rule.CLEAR_AHEAD] = true;
    theRules[9] = new Rule (code, text, setFlags);
    clearFlags ();

    // passing marks and obstructions: not overlapped at the zone
    code = "18.2(c)";
    text =
      "If a boat is clear ahead at the time\n" +
      "she reaches the two-length zone,\n" +
      "the boat clear astern shall\n" +
      "thereafter keep clear. If the boat\n" +
      "clear astern becomes overlapped\n" +
      "outside the other boat she shall\n" +
      "also give the inside boat room. If\n" +
      "the boat clear astern becomes\n" +
      "overlapped inside the other boat\n" +
      "she is not entitled to room. If the\n" +
      "boat that was clear ahead passes\n" +
      "head to wind, Rule 18.2(c) no\n" +
      "longer applies.";
    setFlags[Rule.MARK_RULES] = true;
    setFlags[Rule.INSIDE_TWO_LENGTHS] = true;
    setFlags[Rule.CLEAR_AHEAD] = true;
    theRules[10] = new Rule (code, text, setFlags);
    clearFlags ();

    // passing marks and obstructions: tacking at a mark, forced above
    // close-hauled
    code = "18.3(a)";
    text =
      "If two boats were approaching a\n" +
      "mark on opposite tacks and one\n" +
      "of them completes a tack in the\n" +
      "two-length zone when the other\n" +
      "is fetching the mark, Rule 18.2\n" +
      "does not apply. The boat that\n" +
      "tacked shall not cause the\n" +
      "other boat to sail above close-\n" +
      "hauled to avoid her or prevent\n" +
      "the other boat from passing the\n" +
      "mark.";
    setFlags[Rule.ABOVE_CLOSE_HAULED] = true;
    setFlags[Rule.INSIDE_TWO_LENGTHS] = true;
    setFlags[Rule.MARK_RULES] = true;
    setFlags[Rule.OVERLAPPED] = true;
    setFlags[Rule.TACKING_TO_PORT] = true;
    setFlags[Rule.TACKING_TO_STARBOARD] = true;
    theRules[11] = new Rule (code, text, setFlags);
    clearFlags ();

    // passing marks and obstructions: tacking at a mark, giving room inside
    code = "18.3(b)";
    text =
      "If two boats were approaching a\n" +
      "mark on opposite tacks and one\n" +
      "of them completes a tack in the\n" +
      "two-length zone when the other\n" +
      "is fetching the mark, Rule 18.2\n" +
      "does not apply. The boat that\n" +
      "tacked (b) shall give room if\n" +
      "the other boat becomes overlapped\n" +
      "inside her, in which case Rule 15\n" +
      "does not apply.";
    setFlags[Rule.INSIDE_TWO_LENGTHS] = true;
    setFlags[Rule.MARK_RULES] = true;
    setFlags[Rule.OVERLAPPED] = true;
    setFlags[Rule.TACKING_TO_PORT] = true;
    setFlags[Rule.TACKING_TO_STARBOARD] = true;
    theRules[12] = new Rule (code, text, setFlags);
    clearFlags ();
  }

  // clear the flag array before constructing a new one

  private void clearFlags () {
    int i;

    for (i = 0; i < Rule.NUM_FLAGS; i++)
      setFlags[i] = false;
  }

  // void temporary rules array indexes

  public void clearRules () {
    int i;

    for (i = 0; i < NUM_RULES; i++)
      tempRules[i] = 0;
  }

  // return an array of rule indexes applying for the given flags

  public int[] findRules (boolean inputFlags[]) {
    int t = 0;
    int i, j;
    boolean addProximity = false;

    for (i = 0; i < Rule.NUM_FLAGS; i++)
      if (i == Rule.PROXIMITY && inputFlags[i]) {
       addProximity = true;
        tempFlags[i] = false;
      }
      else
        tempFlags[i] = inputFlags[i];
    clearRules ();
    for (i = 1; i < NUM_RULES; i++) {
      for (j = 0; j < Rule.NUM_FLAGS; j++) {
	if (theRules[i].getFlag (j) == tempFlags[j]) {
	  if (j == Rule.NUM_FLAGS - 1)
	    tempRules[t++] = i;
	}
	else
	  break;
      }
    }
    if (addProximity)
      tempRules[t] = proximityRule;
    return (tempRules);
  }

  // return code for rule index

  public static String getCode (int i) {
    return (theRules[i].getCode ());
  }

  // return text for rule index

  public static String getText (int i) {
    return (theRules[i].getText ());
  }
}
