/*********************************************************************

  RulesResolver.java	(C) Kenneth J. Turner, Mark A. Jennings	01/12/01 V1.3

  This class seperates the rules resolving mechanism from the SailRule class.
  Uses the rulebook flag mechanism to show why and methods resolve who. This
  class will likely be expanded in the future.

  KJT 01/12/01:

    o Minor tidy-up of code

    o Overlap/windward calculation rewritten

    o Various rules fixed

*********************************************************************/

import waba.ui.*;
import waba.fx.*;
import waba.sys.*;
import extra.ui.*;

public class RulesResolver {
  private RuleBook ruleBook;
  private int rules[];				// list of rule indexes
  private SailRule main;			// for some operations

  // constructor

  public RulesResolver () {
    ruleBook = new RuleBook ();
  }

  // overlap calculation method

  private boolean checkOverlap (Boat a, Boat b) {
    Boat boatList[];
    Boat c;
    int i, boatMax;

    if (a.isOverlap (b))
      return true;
    else {
      boatList = main.getBoatList ();
      boatMax = main.getBoatCount ();
      for (i = 0; i < boatMax; i++) {
        c = boatList[i];
	if (c != a && c != b && c.isBetween (a, b) &&
	      a.isOverlap (c) && c.isOverlap (b)) {
          return (true);
	}
      }
      return (false);
    }
  }

  // return a case under open water rules (10, 11, 12, 13) for the given
  // pair of boats

  public Case openWater (Boat a, Boat b) {
    boolean tempFlags[] = new boolean[Rule.NUM_FLAGS];
    int courseRules[] = properCourse (null, a, b); // get proper course rules
    Case c = null;

    boolean overlap = checkOverlap (a, b);
    int aTack = a.getTack ();
    int bTack = b.getTack ();
    boolean aPassedHeadToWind =
      ((a.getState (Boat.TACKING_TO_PORT) && aTack == Boat.PORT) ||
       (a.getState (Boat.TACKING_TO_STARBOARD) && aTack == Boat.STARBOARD));
    boolean bPassedHeadToWind =
      ((b.getState (Boat.TACKING_TO_PORT) && bTack == Boat.PORT) ||
       (b.getState (Boat.TACKING_TO_STARBOARD) && bTack == Boat.STARBOARD));
    boolean aTacking = aPassedHeadToWind && a.isLuffing ();
    boolean bTacking = bPassedHeadToWind && b.isLuffing ();
    boolean bAhead = a.isAhead (b);
    boolean bPort = a.isPort (b);
    boolean bWindward = a.isWindward (b);

    for (int i = 0; i < Rule.NUM_FLAGS; i++)	// init working flags
      tempFlags[i] = false;
    tempFlags[Rule.PROXIMITY] = a.isClose (b);
    rules = null;

    if (aTacking || bTacking) {			// rule 13
      tempFlags[Rule.TACKING_TO_PORT] = true;	// not close-hauled
      tempFlags[Rule.TACKING_TO_STARBOARD] = true;
      tempFlags[Rule.CLEAR_AHEAD] = true;	// has passed
      tempFlags[Rule.HEAD_TO_WIND] = true;	// head to wind
      rules = ruleBook.findRules (tempFlags);
      if (aTacking && bTacking) {		// both tacking?
	if (bPort)				// b is to port of a?
	  c = new Case (a, b, rules);		// a right over b
	else					// a is to port of b
	  c = new Case (b, a, rules);		// b right over a
      }
      else if (aTacking) {			// a tacking?
	c = new Case (b, a, rules);		// b right over a
      }
      else {					// b tacking
	c = new Case (a, b, rules);		// a right over b
      }
    }
    else if (courseRules != null) {		// rule 17
      boolean aAbove = a.getFlag (Boat.ABOVE_PROPER_COURSE);
      boolean bAbove = b.getFlag (Boat.ABOVE_PROPER_COURSE);
      boolean aBelow = a.getFlag (Boat.BELOW_PROPER_COURSE);
      boolean bBelow = b.getFlag (Boat.BELOW_PROPER_COURSE);
      if (aAbove && bAbove) {			// both above course?
	if (bWindward)				// b windward?
	  c = new Case (b, a, courseRules);	// b right over a
	else					// a windward
	  c = new Case (a, b, courseRules);	// a right over b
      }
      else if (aBelow && bBelow) {		// both below course?
	if (bWindward || bAhead)		// b windward/ahead?
	  c = new Case (a, b, courseRules);	// a right over b
	else					// a windward/ahead
	  c = new Case (b, a, courseRules);	// b right over a
      }
      else if (aBelow || aAbove)		// a above/below course?
	c = new Case (b, a, courseRules);	// b right over a
      else					// b above/below course
	c = new Case (a, b, courseRules);	// a right over b
    }
    else {
      if (aTack != bTack) {			// rule 10
	tempFlags[Rule.OPPOSITE_TACKS] = true;
	tempFlags[Rule.STARBOARD_TACK] = true;
	rules = ruleBook.findRules (tempFlags);
	if (aTack == Boat.STARBOARD)		// a is on starboard?
	  c = new Case (a, b, rules);		// a right over b
	else					// a is on port
	  c = new Case (b, a, rules);		// b right over a
      }
      else if (overlap) {			// rule 11
	tempFlags[Rule.OVERLAPPED] = true;
	tempFlags[Rule.TO_LEEWARD] = true;
	rules = ruleBook.findRules (tempFlags);
	if (bWindward)				// b is to windward?
	  c = new Case (a, b, rules);		// a right over b
	else					// a is to windward
	  c = new Case (b, a, rules);		// b right over a
      }
      else if (!overlap) {			// rule 12
	tempFlags[Rule.CLEAR_AHEAD] = true;
	rules = ruleBook.findRules (tempFlags);
	if (bAhead)				// b is ahead?
	  c = new Case (b, a, rules);		// b right over a
	else					// a is ahead
	  c = new Case (a, b, rules);		// a right over b
      }
    }
    return (c);
  }

  // return a list of rule numbers invoked under proper course rules (17)
  // for a given mark (null if there is none) and two boats

  public int[] properCourse (Mark m, Boat a, Boat b) {
    int tempRules[] = null;
    boolean tempFlags[] = new boolean[Rule.NUM_FLAGS];

    boolean near = a.nearBoat (b);
    boolean neitherOnBeat =
      m != null
        ? !(m.onBeat (a) || m.onBeat (b))
	: !(a.isCloseHauled () || b.isCloseHauled ());
    boolean overlapped = a.isOverlap (b);
    boolean sameTack = a.getTack () == b.getTack ();
    boolean aAbove = a.getFlag (Boat.ABOVE_PROPER_COURSE);
    boolean bAbove = b.getFlag (Boat.ABOVE_PROPER_COURSE);
    boolean aBelow = a.getFlag (Boat.BELOW_PROPER_COURSE);
    boolean bBelow = b.getFlag (Boat.BELOW_PROPER_COURSE);
    boolean aGybing = a.getFlag (Boat.GYBING);
    boolean bGybing = b.getFlag (Boat.GYBING);
    boolean aWindward = b.isWindward (a);
    boolean bWindward = a.isWindward (b);

    if (near && sameTack && overlapped && near && // rule 17.1
        ((aAbove && !aWindward) || (bAbove && !bWindward))) {
      tempFlags[Rule.ABOVE_PROPER_COURSE] = true;
      tempFlags[Rule.INSIDE_TWO_LENGTHS] = true;
      tempFlags[Rule.OVERLAPPED] = true;
      tempFlags[Rule.TO_LEEWARD] = true;
      tempRules = ruleBook.findRules (tempFlags);
    }
    else if (near && sameTack && neitherOnBeat && // rule 17.2
	     ((aWindward && aBelow && !aGybing) ||
	      (bWindward && bBelow && !bGybing))) {
      tempFlags[Rule.BELOW_PROPER_COURSE] = true;
      tempFlags[Rule.INSIDE_TWO_LENGTHS] = true;
      tempRules = ruleBook.findRules (tempFlags);
    }
    return (tempRules);
  }

  // for true overlap calculation, a reference to SailRule is needed

  public void registerControl (SailRule s) {
    main = s;
  }

  // return a case with mark rules (18) for the given mark and pair of boats

  public Case withMark (Mark m, Boat a, Boat b) {
    boolean tempFlags[] =			// allocate rule flags
      new boolean[Rule.NUM_FLAGS];
    int courseRules[] = properCourse (m, a, b);	// get proper course rules
    int rules[] = null;
    Case result = openWater (a, b);		// open water rules at first

    for (int i = 0; i < Rule.NUM_FLAGS; i++)	// init working flags
      tempFlags[i] = false;
    tempFlags[Rule.PROXIMITY] = a.isClose (b);

    int aTack = a.getTack ();
    int bTack = b.getTack ();
    boolean bothInZone = a.nearMark (m) && b.nearMark (m);
    boolean bothStarting =
      a.getState (Boat.STARTING) && b.getState (Boat.STARTING) ;
    boolean sameTack = aTack == bTack;
    boolean neitherOnBeat = !(m.onBeat (a) || m.onBeat (b));
    boolean neitherMustTack = !(m.mustTack (a) || m.mustTack (b));

    if (bothInZone && !bothStarting &&		// rule 18
          (sameTack || (neitherOnBeat && neitherMustTack))) {
      tempFlags[Rule.INSIDE_TWO_LENGTHS] = true;
      tempFlags[Rule.MARK_RULES] = true;
      boolean overlapped = checkOverlap (a, b);
      boolean wereOverlapped = a.wasOverlapped (b);
      Boat inside = m.getInside (a, b);
      boolean aAboveCloseHauled = a.aboveCloseHauled ();
      boolean bAboveCloseHauled = b.aboveCloseHauled ();
      boolean aWasAhead = a.getFlag (Boat.WAS_AHEAD);
      boolean bWasAhead = b.getFlag (Boat.WAS_AHEAD);
      boolean aPassedHeadToWind =
        (a.getState (Boat.TACKING_TO_PORT) &&
          aTack == Boat.PORT) ||
        (a.getState (Boat.TACKING_TO_STARBOARD) &&
	  aTack == Boat.STARBOARD);
      boolean bPassedHeadToWind =
        (b.getState (Boat.TACKING_TO_PORT) &&
          bTack == Boat.PORT) ||
        (b.getState (Boat.TACKING_TO_STARBOARD) &&
          bTack == Boat.STARBOARD);
      boolean aForcedUp = aAboveCloseHauled && bPassedHeadToWind;
      boolean bForcedUp = bAboveCloseHauled && aPassedHeadToWind;
      boolean aSqueezedInside = inside == a && bPassedHeadToWind;
      boolean bSqueezedInside = inside == b && aPassedHeadToWind;

      if (sameTack && (aForcedUp || bForcedUp)) { // rule 18.3(a)
        tempFlags[Rule.ABOVE_CLOSE_HAULED] = true;
        tempFlags[Rule.OVERLAPPED] = true;
        tempFlags[Rule.TACKING_TO_PORT] = true;
        tempFlags[Rule.TACKING_TO_STARBOARD] = true;
	rules = ruleBook.findRules (tempFlags);
	if (aForcedUp)				// a forced up by b?
	  result =				// a right over b
	    new Case (b, a, rules, "must not luff");
	else					// b forced up by a
	  result =				// b right over a
	    new Case (a, b, rules, "must not luff");
      }
      else if (sameTack && overlapped &&	// rule 18.3(b)
                 (aSqueezedInside || bSqueezedInside)) {
        tempFlags[Rule.OVERLAPPED] = true;
        tempFlags[Rule.TACKING_TO_PORT] = true;
        tempFlags[Rule.TACKING_TO_STARBOARD] = true;
	rules = ruleBook.findRules (tempFlags);
	if (aSqueezedInside)			// a squeezed inside b?
	  result =				// a right over b
	    new Case (b, a, rules, "must give room to");
	else					// b squeezed inside a
	  result =				// b right over a
	    new Case (a, b, rules, "must give room to");
      }
      else if (aWasAhead && !aPassedHeadToWind) { // rule 18.2(c) for a
        tempFlags[Rule.CLEAR_AHEAD] = true;
	rules = ruleBook.findRules (tempFlags);
	result = new Case (a, b, rules);	// a right over b
      }
      else if (bWasAhead && !bPassedHeadToWind) { // rule 18.2(c) for b
        tempFlags[Rule.CLEAR_AHEAD] = true;
	rules = ruleBook.findRules (tempFlags);
	result = new Case (b, a, rules);	// b right over a
      }
      else if (!overlapped && wereOverlapped) {	// rule 18.2(b)
        tempFlags[Rule.CLEAR_AHEAD] = true;
        tempFlags[Rule.OVERLAPPED] = true;
 	rules = ruleBook.findRules (tempFlags);
	if (inside == a)
	  result = new Case (b, a, rules, "must give room to");
	else
	  result = new Case (a, b, rules, "must give room to");
      }
      else {					// rule 18.2(a)
        tempFlags[Rule.OVERLAPPED] = true;
 	rules = ruleBook.findRules (tempFlags);
	if (inside == a)
	  result = new Case (b, a, rules, "must give room to");
	else
	  result = new Case (a, b, rules, "must give room to");
      }
    }
    result.addToCase (courseRules);
    return (result);
  }
}
