package jgui.ui;

import waba.ui.*;
import waba.fx.*;
import waba.sys.*;
import waba.util.*;
import extra.ui.*;

/**
 * The <code>Scrollbar</code> is used for scrolling the
 * contents of a <code>Scroller</code>.  It allows the user
 * to drag the bar and also to click in the "bed" of the bar
 * to move the bar in jumps.  Arrows at either end of the scrollbar
 * allow the user to move the contents.
 *
 * @version $Header: $
 * @author Jim Guistwite
 */

// Ken Turner <kjt@cs.stir.ac.uk>: SuperWaba 2.0 compatibility and correct
// bar drawing

public class Scrollbar extends Control
{
  Vector listeners;

  Rect forward;
  Image f;

  Rect back;
  Image b;

  Rect bar;

  int min;
  int max;
  int inc;
  int val;
  int extent;
  int orientation;
  boolean dragging = false;
  int dragStart;

  float scaleFactor = 1.0f;

  int timerOp;
  Timer timer;

  static final int VERTICAL = 101;
  static final int HORIZONTAL = 102;

  boolean dragImmediate = false;
  boolean visible = true;

  /**
   * Construct a new scrollbar for either horizontal or vertical
   * orientation.
   * @param orientation use either Scrollbar.VERTICAL or Scrollbar.HORIZONTAL
   */
  public Scrollbar(int orientation) {
    bar = new Rect(0,0,0,0);
    inc = 1;
    min = 0;
    max = 10;
    extent = 1;
    listeners = new Vector(1);
    this.orientation = orientation;

    if (orientation == VERTICAL) {
      f = new Image("down.bmp");
      b = new Image("up.bmp");
    }
    else {
      f = new Image("right.bmp");
      b = new Image("left.bmp");
    }
  }

  /**
   * Set the values for the scrollbar.
   * @param val the value of the scrollbar which ranges between min and max
   * @param extent the total length that this scrollbar is to represent.
   * @param min the minimum value for the scrollbar.
   * @param max the maximum value for the scrollbar.
   */
  public void setValues(int val,
			int extent,
			int min,
			int max) {
    this.min = min;
    this.max = max;
    this.extent = extent;
    this.val = val;
    //System.out.println("values " + newValue + " " + newExtent + " " + newMin + " " + newMax);
    computeScale();
    computeBar(val);
  }

  /**
   * Set the value that should be used to increment or decrement the scrollbar
   * value when the user presses the arrow buttons.
   * @param inc increment value.
   */
  public void setIncrement(int inc) {
    this.inc = (int)(inc*scaleFactor);
  }

  /**
   * Set the value of the scrollbar.  This is similar to setValues but
   * does not require you to reset the extent, min and max.
   * @param val the value of the scrollbar between min and max.
   */
  public void setValue(int val) {
    this.val = val;
    computeScale();
    computeBar(val);
    repaint();
  }

  /**
   * Add a listener to the scrollbar to be notified when the value
   * of the scrollbar changes.
   * @param l listener to be notified when scrollbar changes value.
   */
  public void addScrollbarListener(ScrollbarListener l) {
    listeners.add(l);
  }

  /**
   * Remove a listener from the scrollbar.
   * @param l listener to be notified when scrollbar changes value.
   */
  public void removeScrollbarListener(ScrollbarListener l) {
    listeners.del(listeners.find(l));
  }

  /**
   * Private method to compute the scaleFactor.
   */
  private void computeScale() {
    int diff = max - min;

    if (orientation == VERTICAL) {
      scaleFactor = (float)(height - 10) / (float)diff;
    }
    else {
      scaleFactor = (float)(width - 10) / (float)diff;
    }
  }

  /**
   * Override Control.onEvent to intercept Timer and Pen events.
   * @param e event to be processed.
   */
  public void onEvent(Event e) {
    if (visible == false)
      return;

    //System.out.println(this + " got event " + e);
    if (e.type == ControlEvent.TIMER) {
      if (timerOp >= 10) {
	removeTimer(timer);
	timerOp = timerOp - 10;
	timer = addTimer(250);
      }
      else if (timerOp == 0)
	forward(inc);
      else if (timerOp == 1)
	back(inc);
    }
    else if (e instanceof PenEvent) {
      PenEvent pe = (PenEvent)e;

      if ((pe.type == PenEvent.PEN_UP) && dragging && !dragImmediate) {
	dragging = false;

	// calculate distance moved.
	if (orientation == VERTICAL) {
	  int diff = pe.y - dragStart;
	  if (diff < 0)
	    back((int)(-1*diff/scaleFactor));
	  else
	    forward((int)(diff/scaleFactor));
	  dragStart = 0;
	}
	else {
	  int diff = pe.x - dragStart;
	  if (diff < 0)
	    back((int)(-1*diff/scaleFactor));
	  else
	    forward((int)(diff/scaleFactor));
	  dragStart = 0;
	}
      } else if ((pe.type == PenEvent.PEN_UP) && (timer != null)) {
	removeTimer(timer);
	timer = null;
      }
      else if ((pe.type == PenEvent.PEN_DOWN) && (rectContains(forward,pe.x,pe.y))) {
	forward(inc);
	timerOp = 10;
	timer = addTimer(1000);
      }
      else if ((pe.type == PenEvent.PEN_DOWN) && (rectContains(back,pe.x,pe.y))) {
	back(inc);
	timerOp = 11;
	timer = addTimer(1000);
      }
      else if ((pe.type == PenEvent.PEN_DOWN) && (rectContains(bar,pe.x,pe.y))) {
	//System.out.println("begin drag");
	// begin drag.
	dragging = true;
	if (orientation == VERTICAL)
	  dragStart = pe.y;
	else
	  dragStart = pe.x;
      }
      else if (pe.type == PenEvent.PEN_DOWN) {
	if (orientation == VERTICAL) {
	  if (pe.y > bar.y) {
	    forward(inc*4);
	  }
	  else {
	    back(inc*4);
	  }
	}
	else {
	  if (pe.x > bar.x) {
	    forward(inc*4);
	  }
	  else {
	    back(inc*4);
	  }
	}
      }
      else if (dragging && (pe.type == PenEvent.PEN_DRAG)) {
	//System.out.println("dragging");
	// drag the scrollbar.
	int diff;
	int loc;
	if (orientation == VERTICAL) {
	  loc = pe.y;
	}
	else {
	  loc = pe.x;
	}
	diff = loc - dragStart;
	//System.out.println(loc);

	int proposed = val + (int)(diff/scaleFactor);
	//System.out.println("proposed value: " + proposed);
	if (proposed < min)
	  proposed = min;
	else if ((proposed + extent) > max) {
	  proposed = max - extent;
	}

	if (val != proposed) {
	  if (dragImmediate) {
	    dragStart = loc;
	    fireChange(proposed);
	  }
	  else {
	    computeBar(proposed);
	    Graphics g1 = createGraphics();
            // KJT 20/02/02: set bar clip rectangle
	    g1.setClip (new Rect (0, 0, width, height));
	    onPaint(g1);
	    g1.free();
	  }
	}
      }
    }
  }

  /**
   * Move the scrollbar forward by the amount provided.
   * @param increment move the scrollbar forward by this amount.
   */
  public void forward(int increment) {
    // System.out.println("forward " + increment);
    // System.out.println("sb: " + this);
    if ((val + extent + increment) <= max) {
      fireChange(val+increment);
    }
    else {
      fireChange(max-extent);
    }
  }

  /**
   * Move the scrollbar backward by the amount provided.
   * @param increment move the scrollbar backward by this amount.
   */
  public void back(int increment) {
    //System.out.println("back " + increment);
    if ((val - increment) >= min)
      fireChange(val-increment);
    else {
      if (val != min)
	fireChange(min);
    }
  }

  /**
   * Private method to fire the scrollbar value change event.
   * @param val new scrollbar value.
   */
  private void fireChange(int val) {
    //System.out.println("val is now " + val);
    computeBar(val);
    repaint();

    int old = this.val;
    this.val = val;
    for(int i=0;i<listeners.getCount();i++) {
      ((ScrollbarListener)listeners.get(i)).valueChanged(this,old,val);
    }
  }

  /**
   * Private method to calculate the bar sizes.
   * @param val value of scrollbar.
   */
  private void computeBar(int val) {
    //System.out.println("compute bar at " + val);
    //System.out.println("scale factor: " + scaleFactor);
    if (orientation == VERTICAL) {
      bar.x = 1;
      bar.y = 5 + (int)(scaleFactor*(val-min));
      bar.width = 3;
      bar.height = (int)(scaleFactor*extent);
    }
    else {
      bar.x = 5 + (int)(scaleFactor*(val-min));
      bar.y = 1;
      bar.width = (int)(scaleFactor*extent);
      bar.height = 3;
    }

    //System.out.println(bar.x + " " + bar.y + " " + bar.width + " " + bar.height);
  }

  /**
   * Override setRect in order to recompute the scale and update
   * the scrollbar size.
   * @param x x position.
   * @param y y position.
   * @param width width of the scrollbar.
   * @param height height of the scrollbar.
   */
  public void setRect(int x, int y, int width, int height) {
    //System.out.println(" set rect " + x + " " + y + " " + width + " " + height);
    super.setRect(x,y,width,height);
    computeScale();
    computeBar(val);
    forward = null;
    back = null;
  }

  /**
   * Override onPaint to draw the scrollbar.  This draws the
   * line, rectangle and the arrows at the ends.
   * @param g graphics context to draw with.
   */
  public void onPaint(Graphics g) {
    if (visible == false)
      return;

    if (forward == null) {

      if (orientation == VERTICAL) {
	forward = new Rect(0, height-5, 5, 5);
	back = new Rect(0, 0, 5, 5);
      }
      else {
	forward = new Rect(width-5, 0, 5, 5);
	back = new Rect(0, 0, 5, 5);
      }
    }

   if (orientation == VERTICAL) {
      g.fillRect(1, back.height, 3, forward.y);
      g.drawDots(2, back.height, 2, forward.y);
    }
    else {
      g.fillRect(back.width, 2, forward.x, 2);
      g.drawDots(back.width, 2, forward.x, 2);
    }

    // KJT 20/02/02: added setBackColor
    g.setBackColor(0,0,0);

    g.drawImage(f, forward.x, forward.y);
    g.drawImage(b, back.x, back.y);

    g.fillRect(bar.x, bar.y, bar.width, bar.height);
  }

  /**
   * Private method to determine if a point is in a rect.
   */
  private static boolean rectContains(Rect r, int x, int y) {
    return ((x >= r.x) &&
	    (x <= (r.x + r.width)) &&
	    (y >= r.y) &&
	    (y <= (r.y + r.height)));
  }

  /**
   * Return a string representation of this scrollbar.
   */
  public String toString() {
    return "Scrollbar(" + "," + min + "," + max + "," + inc + "," + extent + "," + val + ")";
  }

  /**
   * Mark this scrollbar as visible or not.  When not visible, the
   * onPaint method simply returns.  This is used by the Scroller when
   * the scrollbars are not needed.
   * @param visible true if the scrollbar should be visible.
   */
  public void setVisible(boolean visible) {
    this.visible = visible;
  }
}
